"""
Conversion functionality for animated stickers.

implements the conversion functionality for the pyrlottie backend. exposing a
public function called `convertAnimated`, which is used to perform the conversion.

The `convertAnimated` function takes the following parameters:
    - swd (Path): The sticker working directory (downloads/packName).
    - _threads (int, optional): The number of threads to pass to ThreadPoolExecutor.
		Defaults to 4.
    - frameSkip (int, optional): Skip n number of frames in the interest of
		optimization with a quality trade-off. Defaults to 1.
    - scale (float, optional): Upscale/downscale the images produced. Intended
		for optimization with a quality trade-off. Defaults to 1.

"""

from __future__ import annotations

import asyncio
from pathlib import Path

import pyrlottie


def convertAnimated(swd: Path, _threads: int = 4, frameSkip: int = 1, scale: float = 1) -> int:
	"""Convert animated stickers to webp, gif and png.

	Args:
	----
		swd (Path): the sticker working directory (downloads/packName)
		threads (int, optional): number of threads to pass to ThreadPoolExecutor. Defaults to 4.
		frameSkip (int, optional): skip n number of frames in the interest of
		optimisation with a quality trade-off. Defaults to 1.
		scale (float, optional): upscale/ downscale the images produced. Intended
		for optimisation with a quality trade-off. Defaults to 1.

	Returns:
	-------
		int: number of stickers successfully converted

	"""
	converted = 0

	def doConvMultLottie(filemaps: list[pyrlottie.FileMap], frameSkip: int, scale: float) -> int:
		return (
			len(
				asyncio.get_event_loop().run_until_complete(
					pyrlottie.convMultLottie(filemaps, frameSkip=frameSkip, scale=scale)
				)
			)
			// 2
		)

	converted += doConvMultLottie(
		filemaps=[
			pyrlottie.FileMap(
				pyrlottie.LottieFile(stckr.absolute().as_posix()),
				{
					stckr.absolute().as_posix().replace("tgs", "gif"),
					stckr.absolute().as_posix().replace("tgs", "webp"),
				},
			)
			for stckr in swd.glob("**/*.tgs")
		],
		frameSkip=frameSkip,
		scale=scale,
	)
	return converted



"""Download sticker packs from Telegram."""

from __future__ import annotations

import argparse
import functools
import operator
from pathlib import Path
from sys import exit as sysexit

from loguru import logger

from tstickers.convert import Backend
from tstickers.downloader import StickerDownloader


def cli() -> None:  # pragma: no cover
	"""Cli entry point."""
	parser = argparse.ArgumentParser("Welcome to TStickers, providing all of your sticker needs")
	parser.add_argument(
		"-t",
		"--token",
		help="Pass in a bot token inline",
	)
	parser.add_argument(
		"-p",
		"--pack",
		action="append",
		nargs="+",
		help="Pass in a pack url inline",
	)
	parser.add_argument(
		"--frameskip",
		default=1,
		type=int,
		help="Set frameskip. default=1",
	)
	parser.add_argument(
		"--scale",
		default=1,
		type=float,
		help="Set scale. default=1.0",
	)
	parser.add_argument(
		"-b",
		"--backend",
		choices=["rlottie-python", "pyrlottie"],
		default="pyrlottie",
		help="Specify the convert backend",
	)
	args = parser.parse_args()

	# Get the token
	token = args.token
	if token is None:
		token = ""
		for candidate in [Path.cwd() / "env.txt", Path.cwd() / "env"]:
			if candidate.exists():
				token = candidate.read_text(encoding="utf-8").strip()
		if not token:
			logger.error(
				'!! Generate a bot token and paste in a file called "env". Send a '
				"message to @BotFather to get started"
			)
			sysexit(1)
	# Get the packs
	packs = functools.reduce(operator.iadd, args.pack or [[]], [])
	if len(packs) == 0:
		while True:
			name = input("Enter sticker_set url (leave blank to stop): ").strip()
			if name == "":
				break
			packs.append(name)
	packs = [name.split("/")[-1] for name in packs]

	downloader = StickerDownloader(token)
	for pack in packs:
		logger.info("=" * 60)
		stickerPack = downloader.getPack(pack)
		if stickerPack is None:
			continue
		logger.info("-" * 60)
		_ = downloader.downloadPack(stickerPack)
		logger.info("-" * 60)

		backend_map = {"rlottie-python": Backend.RLOTTIE_PYTHON, "pyrlottie": Backend.PYRLOTTIE}

		downloader.convertPack(
			pack,
			args.frameskip,
			args.scale,
			backend=backend_map.get(args.backend, Backend.PYRLOTTIE),
		)



"""Sticker download functions used by the module entry point."""

from __future__ import annotations

import time
import urllib.parse
from concurrent.futures import ThreadPoolExecutor, as_completed
from json.decoder import JSONDecodeError
from pathlib import Path
from sys import exit as sysexit
from typing import Any

from emoji import demojize
from loguru import logger

from tstickers import caching
from tstickers.convert import Backend, assure_dir_exists, convertAnimated, convertStatic


class Sticker:
	"""Sticker instance attributes."""

	def __init__(
		self,
		name: str = "None",
		link: str = "None",
		emoji: str = "😀",
		fileType: str = "webp",
	) -> None:
		self.name = name
		self.link = link
		self.emoji = emoji
		self.fileType = fileType

	def __repr__(self) -> str:
		return f"<Sticker:{self.name}>"

	def emojiName(self) -> str:
		"""Get the emoji as a string."""
		return demojize(self.emoji)[1:-1]


class StickerDownloader:
	"""The StickerDownloader sets up the api and makes requests."""

	def __init__(
		self, token: str, session: caching.CachedSession | None = None, multithreading: int = 4
	) -> None:
		self.threads = multithreading
		self.token = token
		self.cwd = assure_dir_exists(Path(), "downloads")
		if session is None:
			self.session = caching.cachedSession
		else:
			self.session = session
		self.api = f"https://api.telegram.org/bot{self.token}/"
		verify = self.doAPIReq("getMe", {})
		if verify is not None and verify["ok"]:
			pass
		else:
			logger.info("Invalid token.")
			sysexit(1)

	def doAPIReq(self, function: str, params: dict[Any, Any]) -> dict[Any, Any] | None:
		"""Use the telegram api.

		Args:
		----
			function (str): function to execute
			params (dict[Any, Any]): function parameters

		Raises:
		------
			RuntimeError: In the event of a failure

		Returns:
		-------
			Optional[dict[Any, Any]]: api response

		"""
		urlParams = "?" + urllib.parse.urlencode(params)
		res = self.session.get(f"{self.api}{function}{urlParams}")
		try:
			res = res.json()
		except JSONDecodeError:
			res = {"ok": False, "raw": res}
		if res["ok"]:
			return res

		logger.info(
			f'API method {function} with params {params} failed. Error: "{res["description"]}"'
		)
		return None

	def getSticker(self, fileData: dict[Any, Any]) -> Sticker:
		"""Get sticker info from the server.

		Args:
		----
			fileData (dict[Any, Any]): sticker id

		Returns:
		-------
			Sticker: Sticker instance

		"""
		info = self.doAPIReq("getFile", {"file_id": fileData["file_id"]})
		if info is not None:
			filePath = info["result"]["file_path"]
			return Sticker(
				name=filePath.split("/")[-1],
				link=f"https://api.telegram.org/file/bot{self.token}/{filePath}",
				emoji=fileData["emoji"],
				fileType=filePath.split(".")[-1],
			)
		return Sticker()

	def getPack(self, packName: str) -> dict[str, Any] | None:
		"""Get a list of File objects.

		Args:
		----
			packName (str): name of the pack

		Returns:
		-------
			dict[str, Any]: dictionary containing sticker data

		"""
		params = {"name": packName}
		res = self.doAPIReq("getStickerSet", params)
		if res is None:
			return None
		stickers = res["result"]["stickers"]
		files = []

		logger.info(f'Starting to scrape "{packName}" ..')
		start = time.time()
		with ThreadPoolExecutor(max_workers=self.threads) as executor:
			futures = [executor.submit(self.getSticker, i) for i in stickers]
			files = [i.result() for i in as_completed(futures)]
		end = time.time()
		logger.info(f"Time taken to scrape {len(files)} stickers - {end - start:.3f}s")
		logger.info("")

		return {
			"name": res["result"]["name"].lower(),
			"title": res["result"]["title"],
			"files": files,
		}

	def downloadSticker(self, path: Path, link: str) -> int:
		"""Download a sticker from the server.

		Args:
		----
			path (Path): the path to write to
			link (str): the url to the file on the server

		Returns:
		-------
			int: path.write_bytes(res.content)

		"""
		return path.write_bytes(self.session.get(link).content)

	def downloadPack(self, pack: dict[str, Any]) -> bool:
		"""Download a sticker pack.

		Args:
		----
			pack (dict[str, Any]): dictionary representing a sticker pack

		Returns:
		-------
			bool: success

		"""
		swd = assure_dir_exists(self.cwd, pack["name"])
		downloads = 0
		logger.info(f'Starting download of "{pack["name"]}" into {swd}')
		start = time.time()
		with ThreadPoolExecutor(max_workers=self.threads) as executor:
			futures = [
				executor.submit(
					self.downloadSticker,
					assure_dir_exists(swd, sticker.fileType)
					/ (
						f'{sticker.name.split("_")[-1].split(".")[0]}+{sticker.emojiName()}'
						f".{sticker.fileType}"
					),
					link=sticker.link,
				)
				for sticker in pack["files"]
			]
			for i in as_completed(futures):
				downloads += 1 if i.result() > 0 else 0
		self.session.close()

		end = time.time()
		logger.info(f"Time taken to download {downloads} stickers - {end - start:.3f}s")
		logger.info("")
		return downloads == pack["files"]

	def convertPack(
		self,
		packName: str,
		frameSkip: int = 1,
		scale: float = 1,
		*,
		noCache: bool = False,
		backend: Backend = Backend.UNDEFINED,
	) -> None:
		"""Convert the webp to gif and png; tgs to gif, webp (webp_animated) and png.

		Args:
		----
			packName (str): name of the directory to convert
			frameSkip (int, optional): skip n number of frames in the interest of
			optimisation with a quality trade-off. Defaults to 1.
			scale (float, optional): upscale/ downscale the images produced. Intended
			for optimisation with a quality trade-off. Defaults to 1.
			noCache (bool, optional): set to true to disable cache. Defaults to False.

		"""
		if not noCache and caching.verify_converted(packName):
			return
		# Make directories
		swd = assure_dir_exists(self.cwd, packName)

		# Convert Stickers
		start = time.time()
		total = len([x for x in Path(swd).glob("**/*") if x.is_file()])

		logger.info(f'Converting stickers "{packName}"...')

		# tgs
		converted = convertedTgs = convertAnimated(
			swd, self.threads, frameSkip=frameSkip, scale=scale, backend=backend
		)

		# webp
		convertedWebp = convertStatic(swd, self.threads)
		converted += convertedWebp

		end = time.time()
		logger.info(f"Time taken to convert {converted} stickers (total) - {end - start:.3f}s")
		logger.info("")

		caching.create_converted(
			packName,
			data={
				"version": 2,
				"info": {
					"packName": packName,
					"frameSkip": frameSkip,
					"scale": scale,
					"swd": swd.as_posix(),
				},
				"converted": {
					"static": convertedWebp,
					"animated": convertedTgs,
					"total": total,
				},
			},
		)



"""Sticker caching functionality used by the downloader."""

from __future__ import annotations

import json
from pathlib import Path
from typing import Any, Callable

from loguru import logger
from requests_cache.session import CachedSession

# requests_cache
cachedSession = CachedSession(
	".cache/tstickers.requests.sqlite",
	backend="sqlite",
	expire_after=60 * 60 * 12,
	allowable_codes=(200,),
	allowable_methods=("GET", "POST"),
)


CACHE_DIR = Path(".cache")
if not CACHE_DIR.exists():
	CACHE_DIR.mkdir()


def verify_converted(pack_name: str) -> bool:
	"""Verify the cache for a packName eg. "DonutTheDog". Uses the cache "version"
	to call the verify function for that version.

	Args:
	----
		pack_name (str): name of the sticker pack eg. "DonutTheDog"

	Returns:
	-------
		bool: if the converted cache has been verified

	"""
	cache = CACHE_DIR / pack_name
	if cache.exists():
		data = json.loads(cache.read_text(encoding="utf-8"))
		verify_func = _get_verify_function(data.get("version", 1))
		if verify_func(data):
			logger.info(f"-> Cache hit for {pack_name}!")
			return True
	logger.info(f"-> Cache miss for {pack_name}!")
	return False


def _verify_converted_v1(data: dict[str, Any]) -> bool:
	"""Verify the cache for a packName using cache data.

	Args:
	----
		data (dict[Path, Any]): packName cache data to verify

	Returns:
	-------
		bool: if the converted cache has been verified

	"""
	return (
		len(list(Path(f"{data['info']['swd']}").glob("**/*"))) > 0
		and data["converted"]["static"] + data["converted"]["animated"]
		>= data["converted"]["total"]
	)


def create_converted(pack_name: str, data: dict) -> None:
	"""Write cache data to a file identified by packName.

	Args:
	----
		pack_name (str): name of the sticker pack eg. "DonutTheDog"
		data (dict): packName cache data to write to cache

	"""
	cache = CACHE_DIR / pack_name
	cache.write_text(json.dumps(data), encoding="utf-8")


def _get_verify_function(version: int) -> Callable[[dict[str, Any]], bool]:
	"""Get the appropriate cache verification function based on version.

	Args:
	----
		version (int): Cache version

	Returns:
	-------
		Callable[[dict[str, Any]], bool]: Cache verification function

	"""
	return {
		1: _verify_converted_v1,
	}.get(version, _verify_converted_v1)



"""Sticker convert functions used by the downloader."""

from __future__ import annotations

import contextlib
import multiprocessing
import time
from concurrent.futures import ThreadPoolExecutor, as_completed
from enum import IntEnum, auto
from pathlib import Path

from loguru import logger
from PIL import Image


class Backend(IntEnum):
	"""Represents different conversion libraries such as pyrlottie, and rlottie-python."""

	UNDEFINED = -1
	PYRLOTTIE = auto()
	RLOTTIE_PYTHON = auto()


def convertAnimatedFunc(_swd: Path, _threads: int, _frameSkip: int, _scale: float) -> int:
	"""Convert animated stickers with (Base/Backend.UNDEFINED)."""
	msg = "Backend could not be loaded"
	raise RuntimeError(msg)


convertRlottiePython = convertPyRlottie = convertAnimatedFunc

with contextlib.suppress(ModuleNotFoundError):
	from tstickers.convert_rlottie_python import convertAnimated as convertRlottiePython
with contextlib.suppress(ModuleNotFoundError):
	from tstickers.convert_pyrlottie import convertAnimated as convertPyRlottie


def assure_dir_exists(*parts: Path | str) -> Path:
	"""Make the directory if it does not exist.

	Args:
	----
		parts (Path): path parts

	Returns:
	-------
		Path: the full path

	"""
	full_path = Path(*parts)
	full_path.mkdir(parents=True, exist_ok=True)
	return full_path


def convertWithPIL(inputFile: str) -> str:
	"""Convert the webp file to png.

	Args:
	----
		inputFile (str): path to input file

	Returns:
	-------
		str: path to input file

	"""
	img = Image.open(inputFile)
	img.save(inputFile.replace("webp", "png"))
	gifImage = inputFile.replace("webp", "gif")
	if not Path(gifImage).exists():
		img.save(gifImage)
	return inputFile


def convertStatic(swd: Path, threads: int = 4) -> int:
	"""Convert static stickers to png and gif.

	Args:
	----
		swd (Path): the sticker working directory (downloads/packName)
		threads (int, optional): number of threads to pass to ThreadPoolExecutor. Defaults to 4.

	Returns:
	-------
		int: number of stickers successfully converted

	"""
	converted = 0
	start = time.time()
	assure_dir_exists(swd, "png")
	assure_dir_exists(swd, "gif")
	with ThreadPoolExecutor(max_workers=threads) as executor:
		for _ in as_completed(
			[
				executor.submit(convertWithPIL, inputFile.as_posix())
				for inputFile in (swd / "webp").glob("**/*")
			]
		):
			converted += 1
	end = time.time()
	logger.info(f"Time taken to convert {converted} stickers (webp) - {end - start:.3f}s")
	logger.info("")
	return converted


def convertAnimated(
	swd: Path,
	threads: int = multiprocessing.cpu_count(),
	frameSkip: int = 1,
	scale: float = 1,
	backend: Backend = Backend.UNDEFINED,
) -> int:
	"""Convert animated stickers to webp, gif and png.

	Args:
	----
		swd (Path): the sticker working directory (downloads/packName)
		threads (int, optional): number of threads to pass to ThreadPoolExecutor. Defaults
			to number of cores/ logical processors.
		frameSkip (int, optional): skip n number of frames in the interest of
		optimisation with a quality trade-off. Defaults to 1.
		scale (float, optional): upscale/ downscale the images produced. Intended
		for optimisation with a quality trade-off. Defaults to 1.
		backend (Backend): The backend to use for conversion. Defaults to Backend.UNDEFINED,
	allowing the system to determine the appropriate library to use.

	Returns:
	-------
		int: number of stickers successfully converted

	"""
	if backend == Backend.UNDEFINED:
		msg = "You must specify a conversion backend"
		raise RuntimeError(msg)
	start = time.time()
	assure_dir_exists(swd, "apng")
	assure_dir_exists(swd, "gif")
	assure_dir_exists(swd, "webp")

	convertMap = {
		Backend.UNDEFINED: convertAnimatedFunc,
		Backend.PYRLOTTIE: convertPyRlottie,
		Backend.RLOTTIE_PYTHON: convertRlottiePython,
	}

	converted = convertMap[backend](swd, threads, frameSkip, scale)

	end = time.time()
	logger.info(f"Time taken to convert {converted} stickers (tgs) - {end - start:.3f}s")
	logger.info("")
	return converted



"""
Conversion functionality for animated stickers.

implements the conversion functionality for the rlottie_python backend. exposing a
public function called `convertAnimated`, which is used to perform the conversion.

The `convertAnimated` function takes the following parameters:
    - swd (Path): The sticker working directory (downloads/packName).
    - _threads (int, optional): The number of threads to pass to ThreadPoolExecutor.
		Defaults to 4.
    - frameSkip (int, optional): Skip n number of frames in the interest of
		optimization with a quality trade-off. Defaults to 1.
    - scale (float, optional): Upscale/downscale the images produced. Intended
		for optimization with a quality trade-off. Defaults to 1.

"""

import concurrent.futures
import multiprocessing
from pathlib import Path

from loguru import logger
from rlottie_python import LottieAnimation


def convert_single_tgs(stckr: Path, fps: int, scale: float = 1.0) -> int:
	tgs_file = stckr.absolute().as_posix()

	# Read the animation outside the context manager to avoid issues with pickling
	anim = LottieAnimation.from_tgs(path=tgs_file)

	try:
		width, height = anim.lottie_animation_get_size()
		fps_orig = anim.lottie_animation_get_framerate()
		fps = min(fps, fps_orig)
		width = int(width * scale)
		height = int(height * scale)

		anim.save_animation(tgs_file.replace("tgs", "apng"), fps, width=width, height=height)
		anim.save_animation(tgs_file.replace("tgs", "gif"), fps, width=width, height=height)
		anim.save_animation(tgs_file.replace("tgs", "webp"), fps, width=width, height=height)

	finally:
		anim.lottie_animation_destroy()

	return 1


def convertAnimated(
	swd: Path,
	threads: int = multiprocessing.cpu_count(),
	frameSkip: int = 1,
	scale: float = 1,
) -> int:
	"""Convert animated stickers to webp, gif and png.

	Args:
	----
		swd (Path): the sticker working directory (downloads/packName)
		threads (int, optional): number of threads to pass to ThreadPoolExecutor. Defaults
			to number of cores/ logical processors.
		frameSkip (int, optional): skip n number of frames in the interest of
		optimisation with a quality trade-off. Defaults to 1.
		scale (float, optional): upscale/ downscale the images produced. Intended
		for optimisation with a quality trade-off. Defaults to 1.

	Returns:
	-------
		int: number of stickers successfully converted

	"""
	converted = 0

	fps = [None, 30, 20, 15, 12][min(4, max(0, frameSkip))]

	with concurrent.futures.ProcessPoolExecutor(max_workers=threads) as executor:
		# Using list comprehension to submit tasks to the executor
		future_to_variable = {
			executor.submit(convert_single_tgs, stckr, fps, scale): stckr
			for stckr in swd.glob("**/*.tgs")
		}

		# Wait for all tasks to complete and retrieve results
		for future in concurrent.futures.as_completed(future_to_variable):
			variable = future_to_variable[future]
			try:
				converted += future.result()
			except Exception as e:
				logger.error(f"Error processing {variable}: {e}")

	return converted



"""Entry point for python -m sigstickers."""

import sys

from loguru import logger

from tstickers.cli import cli

_ = cli


logger.remove(0)
logger.add(
	sys.stderr,
	format="<level>{level: <8}</level> | <level>{message}</level>",
)
